import os
import numpy as np
from copy import copy
from doc2data.utils import load_image

try:
    import tensorflow as tf
    _tf_available = True
except ImportError:
    _tf_available = False

try:
    import torch
    _torch_available = True
except ImportError:
    _torch_available = False

if _tf_available:
    import tensorflow as tf
elif _torch_available:
    import torch
else:
    raise ImportError(
        'Neither TensorFlow nor PyTorch could be loaded. '
        'Please make sure one of the two is installed.'
    )

class DataProcessorMixin:
    """MixIn class for data processing in the training loop."""

    def __init__(
        self,
        instances_df,
        processor = None
    ):
        self.instances_df = instances_df.copy()
        self.processor = processor
        self.inference_mode = None

    def load_instance(self, source):
        pass

    def get_processed_instance(self, source):
        features, labels = self.load_instance(source)
        if self.processor:
            features, labels = self.processor(features, labels, self.inference_mode)

        return features, labels

if _tf_available:

    class DataProcessor(DataProcessorMixin, tf.keras.utils.Sequence):
        """Base data processing pipeline for TensorFlow backend."""

        batch_size = 1

        def __len__(self):
            return int(np.ceil(len(self.instances_df) / self.batch_size))

        def __getitem__(self, idx):
            batch_df = self.instances_df.iloc[idx * self.batch_size:(idx + 1) * self.batch_size]

            batch_features = []
            batch_labels = []
            for i, row in batch_df.iterrows():
                features, labels = self.get_processed_instance(row)
                batch_features.append(features)
                batch_labels.append(labels)
            if isinstance(features, list):
                batch_features = [
                    np.stack([i[0] for i in batch_features], axis = 0),
                    np.stack([i[1] for i in batch_features], axis = 0)
                ]
            else:
                batch_features = np.stack(batch_features, axis = 0)

            batch_labels = np.stack(batch_labels, axis = 0)

            return batch_features, batch_labels

        def on_epoch_end(self):
            pass

        def initialize_split(self, data_split, batch_size, shuffle = None, inference_mode = 'auto'):
            """Returns a Keras Sequence object."""

            sequence = copy(self)
            sequence.instances_df = sequence.instances_df.query('data_split == %s'%str([data_split]))
            sequence.batch_size = batch_size
            if shuffle: print('Shuffle not applicable for tf.keras.Sequence object.')
            if inference_mode == 'auto':
                sequence.inference_mode = False if data_split == 'train_set' else True
            else:
                sequence.inference_mode = inference_mode
            return sequence

if _torch_available and not _tf_available:

    class DataProcessor(DataProcessorMixin, torch.utils.data.Dataset):
        """Base data processing pipeline for PyTorch backend."""

        def __len__(self):
            return len(self.instances_df)

        def __getitem__(self, idx):
            # Note: https://github.com/pytorch/pytorch/issues/13246#issuecomment-905703662
            row = self.instances_df.iloc[idx]
            features, labels = self.get_processed_instance(row)

            return features, labels

        def initialize_split(self, data_split, batch_size, shuffle, inference_mode = 'auto'):
            """Returns a PyTorch DataLoader object."""

            dataset = copy(self)
            dataset.instances_df = dataset.instances_df.query('data_split == %s'%str([data_split]))
            if inference_mode == 'auto':
                dataset.inference_mode = False if data_split == 'train_set' else True
            else:
                dataset.inference_mode = inference_mode
            data_loader = torch.utils.data.DataLoader(
                dataset,
                batch_size = batch_size,
                shuffle = shuffle
            )

            return data_loader



import os
import json
import numpy as np
import pandas as pd
from functools import partial
from tqdm import tqdm
from tqdm.contrib.concurrent import process_map


class Extractor():
    """Class that implements feature extraction from a PDFCollection object."""

    @staticmethod
    def extract_page_image(pdf, target_dir, format, dpi):
        records = []
        for page in pdf.parsed_pages.values():

            image = page.read_contents('page_image', dpi = dpi, force_rgb = True)

            file_name = '%s_page_%s.%s'%(os.path.splitext(pdf.file_name)[0], page.number, format)
            file_path = os.path.join(target_dir, file_name)
            image.save(file_path, quality = 100)

            records.append({
                'file_name': pdf.file_name,
                'page_nr': page.number,
                'image_path': file_path
            })

        return pd.DataFrame(records)

    def extract_page_images_from_pdfs(
        self,
        pdf_collection,
        target_dir,
        format = 'jpg',
        dpi = None,
        overwrite = False
    ):
        os.makedirs(target_dir, exist_ok = overwrite)
        paths = process_map(
            partial(self.extract_page_image, target_dir = target_dir, format = format, dpi = dpi),
            pdf_collection.pdfs.values(),
            chunksize = 1
        )
        pdf_collection.image_path_df = pd.concat(paths, ignore_index = True)

    @staticmethod
    def extract_tokens(pdf, target_dir, ocr_engine, dpi):
        records = []
        for page in pdf.parsed_pages.values():

            if ocr_engine is not None and (page.n_tokens < 5 or page.pcnt_chars_corrupted > 0.05):

                image = page.read_contents('page_image', dpi = dpi, force_rgb = True)
                tokens = ocr_engine(image)
                source = 'ocr'
            else:
                tokens = page.read_contents('tokens')
                source = 'pdf'

            file_name = '%s_page_%s.json'%(os.path.splitext(pdf.file_name)[0], page.number)
            file_path = os.path.join(target_dir, file_name)
            with open(file_path, 'wt') as file:
                json.dump(tokens, file)

            records.append({
                'file_name': pdf.file_name,
                'page_nr': page.number,
                'tokens_path': file_path,
                'tokens_source': source
            })

        return pd.DataFrame(records)

    def extract_tokens_from_pdfs(self, pdf_collection, target_dir, ocr_engine = None, dpi = None, overwrite = False):
        os.makedirs(target_dir, exist_ok = overwrite)
        paths = map(
            partial(self.extract_tokens, target_dir = target_dir, ocr_engine = ocr_engine, dpi = dpi),
            pdf_collection.pdfs.values()
        )
        paths = [i for i in tqdm(paths, total = len(pdf_collection.pdfs.values()))]
        pdf_collection.tokens_path_df = pd.concat(paths, ignore_index = True)

    def generate_spacy_embeddings(self, pdf_collection, target_dir, spacy_nlp, overwrite = False):
        os.makedirs(target_dir, exist_ok = overwrite)

        records = []
        for i, row in tqdm(pdf_collection.tokens_path_df.iterrows(), total = len(pdf_collection.tokens_path_df)):

            # load
            with open(row.tokens_path, 'rt') as file:
                tokens = json.load(file)

            # process
            token_sequence = [i['text'] for i in tokens]
            token_sequence = ' '.join(token_sequence)
            nlp_content = spacy_nlp(token_sequence)

            # write
            file_name = '%s_page_%s.npy'%(os.path.splitext(row.file_name)[0], row.page_nr)
            file_path = os.path.join(target_dir, file_name)
            embedding_sequence = np.array([token.vector for token in nlp_content])
            embedding_sequence.shape
            np.save(file_path, embedding_sequence)

            # record
            records.append({
                'file_name': row.file_name,
                'page_nr': row.page_nr,
                'embeddings_path': file_path,
                'n_words': len(tokens),
                'n_spacy_tokens': len(nlp_content),
                'n_valid_embeddings': sum([i.has_vector for i in nlp_content])
            })

        records = pd.DataFrame(records)
        records['pcnt_identified'] = np.round(records.n_valid_embeddings / records.n_spacy_tokens, 2)
        pdf_collection.embeddings_path_df = records



# SPDX-FileCopyrightText: 2022-present Sergej Levich <sergej.levich@gmail.com>
#
# SPDX-License-Identifier: Apache-2.0

"""doc2data - Integrated document processing with machine learning."""

__version__ = "0.2.0"



import numpy as np
from PIL import Image, ImageDraw
from matplotlib import pyplot as plt
from doc2data.utils import denormalize_bbox

def draw_bounding_boxes(image, bounding_boxes, bbox_color = 'red'):
    """Draws bounding boxes."""

    if isinstance(image, np.ndarray):
        image = Image.fromarray(image)

    image_copy = image.copy()

    # check if only one bounding box was provided
    if isinstance(bounding_boxes[0], float):
        bounding_boxes = [bounding_boxes]

    bounding_boxes = np.array(bounding_boxes)

    for box in bounding_boxes:
        print(box)
        if all(box <= 1):
            box = denormalize_bbox(box, image.width, image.height)
            # maybe -1 on the end_x & end_y coordinate

        draw = ImageDraw.Draw(image_copy)
        draw.rectangle(box, outline = bbox_color)

    return image_copy

def plot_keras_history(history):
    """Generates plot from a Keras history object."""

    fig, (ax1, ax2) = plt.subplots(1, 2, sharex = True)
    fig.suptitle('Training progress')
    ax1.plot(history.history['loss'])
    ax1.plot(history.history['val_loss'])
    ax1.set(title = 'loss')
    ax1.legend(['training', 'validation'], loc='upper right')
    if 'mae' in history.history.keys():
        ax2.plot(history.history['mae'])
        ax2.plot(history.history['val_mae'])
        ax2.set(title = 'mae')
    if 'accuracy' in history.history.keys():
        ax2.plot(history.history['accuracy'])
        ax2.plot(history.history['val_accuracy'])
        ax2.set(title = 'accuracy')
    fig.tight_layout()
    fig.set_size_inches(w = 15, h = 5)
    return fig

# copied from:
# https://github.com/keras-team/keras/blob/v2.9.0/keras/utils/np_utils.py#L21-L74
def to_categorical(y, num_classes=None, dtype='float32'):
  """Converts a class vector (integers) to binary class matrix.
  E.g. for use with `categorical_crossentropy`.
  Args:
      y: Array-like with class values to be converted into a matrix
          (integers from 0 to `num_classes - 1`).
      num_classes: Total number of classes. If `None`, this would be inferred
        as `max(y) + 1`.
      dtype: The data type expected by the input. Default: `'float32'`.
  Returns:
      A binary matrix representation of the input. The class axis is placed
      last.
  Example:
  >>> a = tf.keras.utils.to_categorical([0, 1, 2, 3], num_classes=4)
  >>> a = tf.constant(a, shape=[4, 4])
  >>> print(a)
  tf.Tensor(
    [[1. 0. 0. 0.]
     [0. 1. 0. 0.]
     [0. 0. 1. 0.]
     [0. 0. 0. 1.]], shape=(4, 4), dtype=float32)
  >>> b = tf.constant([.9, .04, .03, .03,
  ...                  .3, .45, .15, .13,
  ...                  .04, .01, .94, .05,
  ...                  .12, .21, .5, .17],
  ...                 shape=[4, 4])
  >>> loss = tf.keras.backend.categorical_crossentropy(a, b)
  >>> print(np.around(loss, 5))
  [0.10536 0.82807 0.1011  1.77196]
  >>> loss = tf.keras.backend.categorical_crossentropy(a, a)
  >>> print(np.around(loss, 5))
  [0. 0. 0. 0.]
  """
  y = np.array(y, dtype='int')
  input_shape = y.shape
  if input_shape and input_shape[-1] == 1 and len(input_shape) > 1:
    input_shape = tuple(input_shape[:-1])
  y = y.ravel()
  if not num_classes:
    num_classes = np.max(y) + 1
  n = y.shape[0]
  categorical = np.zeros((n, num_classes), dtype=dtype)
  categorical[np.arange(n), y] = 1
  output_shape = input_shape + (num_classes,)
  categorical = np.reshape(categorical, output_shape)
  return categorical



# SPDX-FileCopyrightText: 2022-present Sergej Levich <sergej.levich@gmail.com>
#
# SPDX-License-Identifier: Apache-2.0

"""Reading PDF files and creating collections."""

import os
import pickle
import logging
import numpy as np
import pandas as pd
import fitz
from tqdm import tqdm
from tqdm.contrib.concurrent import process_map
from PIL import Image, ImageDraw

from doc2data.utils import (
    get_pcnt_chars_corrupted,
    normalize_bbox,
    denormalize_bbox,
    convert_to_rgb,
)

logger = logging.getLogger(__name__)


class Page:

    """Class representing individual pages from a pdf file.

    This class allows accessing the contents of a page using the pymupdf package.
    It also contains additional attributes that describe the page content.
    All attribute values are obtained using the pymupdf interface therefore relying
    on its accuracy.

    Attributes:
        pt_height: Height of the page in points (1 point = 1/72 inch).
        pt_width: Width of the page in points (1 point = 1/72 inch).
        number: Page number within the pdf file.
        content_type: Type of information contained in the page. One of:

            - text: Pure text.
            - images: One or more images.
            - mixed: A combination of text and images.
            - recovered_image: If pymupdf does not detect any content, the page is
                rendered as an image.
        n_tokens: Number of tokens identified by pymupdf. Roughly corresponds to words.
        n_images: Number of images identified by pymupdf.
        pcnt_chars_corrupted: Number of characters that could not be correctly decoded.
        path_to_pdf: Path to pdf file containing the page.
    """

    def __init__(
        self,
        pt_height,
        pt_width,
        number,
        content_type,
        n_tokens,
        n_images,
        pcnt_chars_corrupted,
        path_to_pdf,
    ):
        self.pt_height = pt_height
        self.pt_width = pt_width
        self.number = number
        self.content_type = content_type
        self.n_tokens = n_tokens
        self.n_images = n_images
        self.pcnt_chars_corrupted = pcnt_chars_corrupted
        self.path_to_pdf = path_to_pdf

    def read_contents(self, types, force_rgb=None, dpi=None):

        """Read page contents via the pymupdf interface.

        This opens the pdf file with pymupdf and extracts the requested
        content. Multiple content types can be provided simultaneously.

        Args:
          types:
            String or list of strings indicating which contents should be
            returned. Possible content types are:

              - tokens: Tokens with bounding boxes. A token rougly corresponds
                  to a word.
              - text: String containg all tokens in the reading order which
                  is recovered by pymupdf.
              - images: Images with bounding boxes.
              - page_image: Entire page as one image.
              - raw_dict: Raw output from pymupdf.
          force_rgb:
            Only if 'page_image' in types: Coverts to RGB and adds white backgound.
          dpi:
            Only if 'page_image' in types: Resolution to use when converting pdf to image.

        Returns:
          A dictionary with the requested contents. If a single type was requested,
            it is returned directly.

        Examples:
          >>> from doc2data.pdf import PDFFile
          >>> pdf_file = PDFFile('path_to_pdf')
          >>> pdf_file.parse_pages()
          >>> page = pdf_file.processed_pages[0]
          >>> page.read_contents('page_image')
          >>> page.read_contents(['tokens', 'images'])

        Raises:
          ValueError: If one or more requested types are not recognized.
        """

        allowed_types = set(("tokens", "text", "images", "page_image", "raw_dict"))
        if isinstance(types, str):
            types = [types]
        if set(types).difference(allowed_types) != set():
            raise ValueError("Content type not recognized")

        with fitz.open(self.path_to_pdf) as pdf:
            page = pdf.load_page(self.number)
            contents = {}
            if "tokens" in types:
                records = []
                words = page.get_text(option="words")
                words.sort(key=lambda x: x[5:8])
                for i, word in enumerate(words):
                    records.append(
                        {
                            "id": i,
                            "bbox": normalize_bbox(
                                word[0:4], self.pt_width, self.pt_height
                            ),
                            "text": word[4],
                        }
                    )
                contents["tokens"] = records
            if "text" in types:
                contents["text"] = page.get_text(option="text")
            if "images" in types:
                records = []
                image_references = page.get_images(full=True)
                for i, ref in enumerate(image_references):
                    pix = fitz.Pixmap(pdf.extract_image(ref[0])["image"])
                    if ref[1] > 0:  # add /SMask or /Mask if exists
                        mask = fitz.Pixmap(pdf.extract_image(ref[1])["image"])
                        pix = fitz.Pixmap(pix, mask)
                    mode = "RGBA" if pix.alpha else "RGB"
                    image = Image.frombytes(mode, [pix.width, pix.height], pix.samples)
                    bbox = page.get_image_bbox(ref)
                    normalized_bbox = normalize_bbox(
                        list(bbox), self.pt_width, self.pt_height
                    )
                    records.append({"id": i, "bbox": normalized_bbox, "image": image})
                contents["images"] = records
            if "page_image" in types:
                pix = page.get_pixmap(alpha=True, dpi=dpi)
                mode = "RGBA" if pix.alpha else "RGB"
                image = Image.frombytes(mode, [pix.width, pix.height], pix.samples)
                if image.mode == "RGBA" and force_rgb:
                    image = convert_to_rgb(image, add_white_background=True)
                contents["page_image"] = image
            if "raw_dict" in types:
                contents["raw_dict"] = page.get_text(option="rawdict")

        if len(types) == 1:
            contents = contents.pop(types[0])

        return contents

    def show_page_image(self, dpi=None, show_bboxes=False, bbox_color="red"):
        """Wrapper for Page.read_contents to show page as image.

        Args:
          dpi: Resolution to use when converting pdf to image.
          show_bboxes: Boolean indicating whether to draw bounding boxes aroung
            the tokens and images.
          bbox_color: Color of the bounding boxes.

        Returns:
            A PIL image of the document page.
        """

        contents = self.read_contents(
            ["page_image", "tokens", "images"], force_rgb=True, dpi=dpi
        )
        image = contents["page_image"]
        if show_bboxes:
            elements = contents["tokens"]
            elements.extend(contents["images"])
            draw = ImageDraw.Draw(image)
            for element in elements:
                scaled_bbox = denormalize_bbox(
                    element["bbox"], image.width, image.height
                )
                draw.rectangle(scaled_bbox, outline=bbox_color)

        return image


class PDFFile:
    """Class representing an individual pdf file.

    Attributes:
      path_to_pdf: Path to pdf file.
      file_name: File name.
      n_pages: Number of pages in the pdf file.
      parsed_pages: List of Page objects providing an interface to each page.
      loaded_successfully: Boolan indicating if pdf file could be processed.
    """

    def __init__(self, path_to_pdf):
        self.path_to_pdf = path_to_pdf
        self.file_name = os.path.basename(path_to_pdf)
        self.n_pages = None
        self.parsed_pages = None
        self.loaded_successfully = None

    def __getitem__(self, item):
        return self.parsed_pages[item]

    def __repr__(self):
        return (
            f"<{self.__module__}.{self.__class__.__name__}"
            f" with {self.n_pages} pages at {hex(id(self))}>"
        )

    def open_fitz_document(self):
        """Opens the pdf file as fitz object."""
        assert self.file_name.lower().endswith(
            ".pdf"
        ), f"File {self.file_name} is not a PDF and will be ignored"
        return fitz.open(filename=self.path_to_pdf)

    def parse_pages(self):
        """Iterates over pages and instantiates Page objects.

        Raises:
          AssertionError: If the file does not have a .pdf extension.
          RuntimeError: If pymupdf fails to open the file.
        """

        try:
            with self.open_fitz_document() as pdf:
                self.n_pages = pdf.page_count
                self.parsed_pages = {}
                for i, page in enumerate(pdf):
                    page_content = page.get_text("rawdict")
                    block_types = np.array([b["type"] for b in page_content["blocks"]])
                    if len(block_types) == 0:
                        content_type = "recovered_image"
                    elif all(block_types == 0):
                        content_type = "text"
                    elif all(block_types == 1):
                        content_type = "images"
                    else:
                        content_type = "mixed"

                    parsed_page = Page(
                        pt_height=page_content["height"],
                        pt_width=page_content["width"],
                        number=page.number,
                        content_type=content_type,
                        n_tokens=len(page.get_text_words()),
                        n_images=sum(block_types),
                        pcnt_chars_corrupted=get_pcnt_chars_corrupted(page),
                        path_to_pdf=self.path_to_pdf,
                    )

                    self.parsed_pages[i] = parsed_page
                self.loaded_successfully = True

        except AssertionError as assertion_error:
            self.loaded_successfully = False
            logger.warning(assertion_error)
        except RuntimeError:  # check for errors raised by pymupdf
            self.loaded_successfully = False
            logger.warning(
                "File %s could not be read and will be ignored", self.file_name
            )


class PDFCollection:
    """Creates a collection from pdf files that are stored in a directory.

    This class serves multiple purposes: First, it allows to parse pdf files
    from a directory to create a collection. Second, it provides overview
    information about the collection on page level. Third, it allows reading
    contents from individual pdfs through a dictionary with PDFFile objects.

    The PDFCollection class does not store any files itself. Instead, it only
    keeps track of the files present in the target folder via a dictionary.
    Therefore, once a collection is created, files should not be removed from
    the source folder.

    Examples:
      >>> from doc2data.pdf import PDFCollection
      >>> pdf_collection = PDFCollection('path_to_files') # create collection
      >>> pdf_collection.parse_files() # populate collection
      >>> print(pdf_collection.overview) # inspect collection

    Attributes:
      path_to_files: Path to source directory containing the pdf files.
      pdfs: A dictionary containing a PDFFile object for each file.      
      ignored_files: A list of file names that could not be processed.
      overview: A Pandas DataFrame containing summary information on page level.
    """

    def __init__(self, path_to_files):
        self.path_to_files = path_to_files
        self.pdfs = None
        self.ignored_files = None
        self.overview = None

    def count_source_files(self):
        """Counts files in the source directory."""
        return len(os.listdir(self.path_to_files))

    def parse_single_file(self, file_name):
        """Instantiates PDFFile object and parses pages."""

        pdf = PDFFile(os.path.join(self.path_to_files, file_name))
        pdf.parse_pages()
        return pdf

    def parse_files(self, use_multiprocessing=True):
        """Populates the collection with PDFFile objects based on source directory.

        Additionally, information on the content is extracted and recorded on
        page level.

        Args:
          use_multiprocessing: Use multiprocessing to speed up population using
            all available cores
        """

        logger.info("Parsing files in %s", self.path_to_files)

        all_files = os.listdir(self.path_to_files)

        if use_multiprocessing:
            pdf_list = process_map(self.parse_single_file, all_files, chunksize=10)
        else:
            pdf_list = list(map(self.parse_single_file, tqdm(all_files)))

        self.pdfs = {i.file_name: i for i in pdf_list if i.loaded_successfully}
        self.ignored_files = [
            i.file_name for i in pdf_list if not i.loaded_successfully
        ]

        overview = []
        for pdf in self.pdfs.values():
            if pdf.loaded_successfully:
                for page in pdf.parsed_pages.values():
                    overview.append(
                        {
                            "file_name": pdf.file_name,
                            "page_nr": page.number,
                            "pt_height": page.pt_height,
                            "pt_width": page.pt_width,
                            "size": page.pt_height * page.pt_width,
                            "content_type": page.content_type,
                            "n_tokens": page.n_tokens,
                            "n_images": page.n_images,
                            "pcnt_chars_corrupted": page.pcnt_chars_corrupted,
                        }
                    )
        self.overview = pd.DataFrame(overview)

        logger.info(
            "%s files parsed, %s files ignored", len(all_files), len(self.ignored_files)
        )

    def save(self, file_path, overwrite=False):
        """Serializes collection to a file with pickle.

        Path directories are created if they do not exist.

        Args:
          file_path: File path where to save the collection.
          overwrite: Set to True if an existing collection should be overwritten.
        """
        if os.path.exists(file_path) and not overwrite:
            raise FileExistsError("Collection already exists")
        if os.path.dirname(file_path):
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "wb") as file:
            pickle.dump(self, file)

    @staticmethod
    def load(file_path):
        """Loads serialized collection.

        Args:
          file_path: File path from where to load the collection

        Returns:
          A PDFCollection object.
        """

        with open(file_path, "rb") as file:
            dataset = pickle.load(file)

        return dataset



import numpy as np
from doctr.models import ocr_predictor

ocr_model = ocr_predictor(det_arch='db_resnet50', reco_arch='crnn_vgg16_bn', pretrained = True)

def run_doctr_ocr(image):
    "Runs docTR OCR on a single image."
    result = ocr_model([np.array(image)])
    page = result.export()['pages'][0]
    tokens = []
    c = 0
    for b in page['blocks']:
        for l in b['lines']:
            for w in l['words']:
                tokens.append({
                    'id': c,
                    'bbox': w['geometry'][0] + w['geometry'][1],
                    'text': w['value']
                })
                c += 1
    return tokens



import os
import numpy as np
import tensorflow as tf
from doc2data.experimental.utils import plot_keras_history
from doc2data.experimental.task_processors import PageRotationClsProcessor

try:
   import mlflow
except:
   pass


class KerasTrainer:
    """Class for training Keras models."""

    @staticmethod
    def train_model(
        training_params,
        model,
        data_processor,
        loss,
        metric,
        mlflow_tracking = False
    ):
        # start mlflow run
        if mlflow_tracking: mlflow.start_run()

        train_data = data_processor.initialize_split(data_split = 'train_set', batch_size = training_params['batch_size'])
        val_data = data_processor.initialize_split(data_split = 'val_set', batch_size = training_params['batch_size'])
        test_data = data_processor.initialize_split(data_split = 'test_set', batch_size = training_params['batch_size'])

        data_processor.train_set = train_data
        data_processor.val_set = val_data
        data_processor.test_set = test_data

        checkpointer = tf.keras.callbacks.ModelCheckpoint(
            'tmp_checkpoints/model.hdf5',
            monitor='val_loss',
            verbose=1,
            save_best_only=True
        )

        model.compile(
            optimizer = tf.keras.optimizers.Adam(learning_rate = training_params['learning_rate']),
            loss = loss,
            metrics = [metric]
        )

        history = model.fit(
            x = train_data,
            epochs = training_params['epochs'],
            validation_data = val_data,
            callbacks = [checkpointer],
            workers = 8,
            use_multiprocessing = True,
            verbose = 1
        )

        model.load_weights(checkpointer.filepath)

        # log results in mlflow
        if mlflow_tracking:
            mlflow.log_params(training_params)
            best_epoch = np.argmin(history.history['val_loss'])
            mlflow.log_metric('best_epoch', best_epoch)
            mlflow.log_metric('best_val_loss', history.history['val_loss'][best_epoch])
            mlflow.log_metric(f'best_val_{metric}', history.history[f'val_{metric}'][best_epoch])
            for key, values in history.history.items():
                for epoch, value in enumerate(values):
                    mlflow.log_metric(key, value, step = epoch)
            plot = plot_keras_history(history)
            mlflow.log_figure(plot, 'training_progress.png')
            os.makedirs('models', exist_ok = True)
            run = mlflow.active_run()
            model.save(os.path.join('models', '%s.h5'%run.info.run_id))
            test_loss, test_metric = model.evaluate(test_data)
            mlflow.log_metrics({'test_loss': test_loss, f'test_{metric}': test_metric})

            # end mlflow run
            mlflow.end_run()



# SPDX-FileCopyrightText: 2022-present Sergej Levich <sergej.levich@gmail.com>
#
# SPDX-License-Identifier: Apache-2.0

"""Utilities."""

import numpy as np
from PIL import Image


def convert_to_rgb(image, add_white_background=True):
    """Converts a PIL Image from RGBA to RGB.

    White background is added per default to prevent transparent pixels to be
    rendered black.
    """

    if image.mode == "RGBA":
        if add_white_background:
            white_bg = Image.new("RGB", image.size, (255, 255, 255))
            white_bg.paste(image, mask=image.split()[3])
            image = white_bg
        else:
            image = image.convert("RGB")

    return image


def load_image(file_path, target_size=None, to_array=True, force_rgb=True):
    """Loads image from file."""

    image = Image.open(file_path)
    if target_size:
        image = image.resize(target_size, resample=None)
    if image.mode == "RGBA" and force_rgb:
        image = convert_to_rgb(
            image, add_white_background=True
        )  # adds white background to RGBA images
    if to_array:
        image = np.array(image)

    return image


def get_pcnt_chars_corrupted(page):
    """Calculates proportion of replacement characters on page."""

    checks = []
    for block in page.get_text("rawdict", flags=0)[
        "blocks"
    ]:  # flags value excludes any images on page
        for line in block["lines"]:
            for span in line["spans"]:
                for char in span["chars"]:
                    checks.append(bool(ord(char["c"]) == 65533))

    pcnt_chars_corrupted = sum(checks) / len(checks) if len(checks) > 0 else None

    return pcnt_chars_corrupted


def normalize_bbox(bounding_box, width, height):
    """Calculates relative coordinates of a bounding box."""

    return [
        bounding_box[0] / width,
        bounding_box[1] / height,
        bounding_box[2] / width,
        bounding_box[3] / height,
    ]


def denormalize_bbox(bounding_box, width, height):
    """Calculates absolute coordinates of a bounding box."""

    return [
        int(bounding_box[0] * width),
        int(bounding_box[1] * height),
        int(bounding_box[2] * width),
        int(bounding_box[3] * height),
    ]



import numpy as np
from doc2data.utils import load_image
from doc2data.experimental.base_processors import DataProcessor
from doc2data.experimental.utils import to_categorical

class PageRotationClsProcessor(DataProcessor):
    """Processor for page rotation."""

    def __init__(
        self,
        instances_df,
        processor = None,
        image_size = (224, 224)
    ):
        super().__init__(instances_df, processor)

        self.instances_df['label_string'] = self.instances_df.labels
        self.instances_df.labels, label_index = self.instances_df.labels.factorize(sort = True)
        self.label_dict = {k: v for k, v in enumerate(label_index)}
        self.n_classes = len(self.label_dict)
        self.image_size = image_size

    def load_instance(self, source):
        features = load_image(source.image_path, target_size = self.image_size)
        labels = source.labels
        return features, labels

    def get_processed_instance(self, source):
        features, labels = super().get_processed_instance(source)
        labels = to_categorical(labels, num_classes = self.n_classes)
        return features, labels


class PageCroppingRegProcessor(DataProcessor):
    """Processor for page cropping."""

    def __init__(
        self,
        instances_df,
        processor = None,
        image_size = (224, 224)
    ):
        super().__init__(instances_df, processor)
        self.image_size = image_size

    def load_instance(self, source):
        features = load_image(source.image_path, target_size = self.image_size)
        labels = np.array(source.labels)
        return features, labels

class BBoxClsProcessor(DataProcessor):
    """Processor for bounding box classification."""

    def __init__(
        self,
        instances_df,
        processor = None,
        image_size = (224, 224)
    ):
        super().__init__(instances_df, processor)

        self.instances_df['label_string'] = self.instances_df.labels
        self.instances_df.labels, label_index = self.instances_df.labels.factorize(sort = True)
        self.label_dict = {k: v for k, v in enumerate(label_index)}
        self.n_classes = len(self.label_dict)
        self.image_size = image_size

    def load_instance(self, source):
        features = np.array(source.bbox)
        labels = to_categorical(source.labels, num_classes = self.n_classes)
        return features, labels

class DocumentClsProcessor(DataProcessor):
    """Processor for document classification."""

    def __init__(
        self,
        instances_df,
        modalities,
        processor = None,
        image_size = (224, 224)
    ):
        super().__init__(instances_df, processor)
        self.instances_df['label_string'] = self.instances_df.labels
        self.instances_df.labels, label_index = self.instances_df.labels.factorize(sort = True)
        self.label_dict = {k: v for k, v in enumerate(label_index)}
        self.n_classes = len(self.label_dict)
        self.max_n_pages = np.max([len(i) for i in instances_df.image_paths])
        self.modalities = modalities
        self.image_size = image_size

    def load_instance(self, source):
        instance_features = []
        if 'image' in self.modalities:
            if isinstance(source.image_paths, list):
                features = []
                for path in source.image_paths:
                    page_image = load_image(path, target_size = self.image_size, to_array = True)
                    features.append(page_image)
                for i in range(self.max_n_pages - len(features)):
                    features.append(np.zeros((self.image_size + (3,))))
                features = np.concatenate(features, axis = 0)
            else:
                features = load_image(source.image_paths, self.image_size)
            instance_features.append(features)
        if 'embeddings' in self.modalities:
            if isinstance(source.embeddings_paths, list):
                features = []
                for path in source.embeddings_paths:
                    page_embeddings = np.load(path)
                    if len(page_embeddings) == 0:
                        # Zero-length embeddings?
                        page_embeddings = np.zeros((1, 300))
                    if len(page_embeddings) > 500:
                        page_embeddings = page_embeddings[0:500,]
                    page_embeddings = np.pad(page_embeddings, ((0, 500 - page_embeddings.shape[0]), (0, 0)))
                    features.append(page_embeddings)
                for i in range(self.max_n_pages - len(features)):
                    features.append(np.zeros((500, 300)))
            else:
                features = np.load(path)
            features = np.concatenate(features, axis = 0)
            instance_features.append(features)

        if len(instance_features) == 1:
            instance_features = instance_features[0]
        label = to_categorical(np.array(source.labels), num_classes = self.n_classes)
        return instance_features, label


class TokenClsProcessorLayoutXLM(DataProcessor):
    """Processor for token classification (LayoutXLM fine-tuning)."""

    def __init__(
        self,
        instances_df,
        processor,
        pdf_collection
    ):
        super().__init__(instances_df, processor)
        self.pdf_collection = pdf_collection
        unique_labels = np.unique([i for j in instances_df.labels for i in j])
        self.id2label = {v: k for v, k in enumerate(unique_labels)}
        self.label2id = {k: v for v, k in enumerate(unique_labels)}

    @staticmethod
    def normalize_bbox_layoutxlm(bbox, width, height):
         return [
             int(1000 * (bbox[0] / width)),
             int(1000 * (bbox[1] / height)),
             int(1000 * (bbox[2] / width)),
             int(1000 * (bbox[3] / height)),
         ]

    def load_instance(self, source):
        pdf = self.pdf_collection.pdfs[source.file_name]
        image = pdf.parsed_pages[0].read_contents('page_image', force_rgb = True)
        tokens = source.tokens
        bboxes = source.bboxes
        bboxes = [self.normalize_bbox_layoutxlm(i, image.width, image.height) for i in bboxes]
        labels = [self.label2id[i] for i in source.labels]

        return (image, tokens, bboxes), labels

    def get_processed_instance(self, source):
        features, labels = self.load_instance(source)
        if self.processor:
            instance = self.processor(features, labels, self.inference_mode)
        else:
            instance = (features, labels)

        return instance

    def __getitem__(self, idx):
        row = self.instances_df.iloc[idx]
        instance = self.get_processed_instance(row)

        return instance
